package lazabs

import java.io.{FileInputStream,InputStream,FileNotFoundException}
import parser._
import lazabs.ast._
import lazabs.ast.ASTree._
import lazabs.prover._
import lazabs.viewer._
import lazabs.cfg.MakeCFG
import lazabs.utils.Manip._
import lazabs.vcg.VCG

object Main {
  var in: InputStream = null
  var fileName = ""
  var funcName = "sc_main"
  var drawCFG = false
  var prettyPrint = false
  var verify = false
  var sign = false
  var log = false
  def readInputFile {
    if( fileName == "") {
      println("Input file missing")
      sys.exit(0)
    }
    try {
      in = new FileInputStream(fileName)
    } catch {
      case e: FileNotFoundException => println("No such file: " + fileName + ". Use -h for usage information" )
                       sys.exit(0)
    }
  }
  def getASTree = {
    val lexer = new Lexer(in)
    val parser = new Parser(lexer)
    val tree = parser.parse()
    if(!(tree.value.isInstanceOf[lazabs.ast.ASTree.Sobject])) {
      throw new RuntimeException("The input file is invalid")
    }
    tree.value.asInstanceOf[lazabs.ast.ASTree.Sobject]
  }

  def main(args: Array[String]) {   
    def arguments(args: List[String]): Unit = args match {
      case Nil =>
      case "-c" :: rest => drawCFG = true; arguments(rest)
      case "-p" :: rest => prettyPrint = true; arguments(rest)
      case "-v" :: rest => verify = true; arguments(rest)
      case "-s" :: rest => sign = true; arguments(rest)
      case mFuncName :: rest if (mFuncName.startsWith("-m:")) => funcName = "sc_" + mFuncName.drop(3); arguments(rest)
      case fn :: rest => fileName = fn;  arguments(rest)
    }
    arguments(args.toList)
    readInputFile
    val ast = getASTree
    if(prettyPrint) {println(ScalaPrinter(ast)); return}
    val cfg = MakeCFG(ast,funcName)
    if(drawCFG) {DrawGraph(cfg)}
    if(sign) {
      import lazabs.ai.AnalysisAlgoritm
      import lazabs.ai.Signs
      val lattice = new Signs.PSignLattice(cfg.variables.toSet)
      val a = new AnalysisAlgoritm(Signs.SignTransFun, lattice, cfg)
      a.computeFixpoint
      a.facts foreach { case (n,fs) => println("Node: " + n + "; Facts: " + fs) }
    }
    if (verify) {
      val vcgs = VCG(cfg) 
      var verified = true
      vcgs foreach { e => 
        println("VCG:") 
        println(ScalaPrinter(e))
        println("Validity:") 
        Prover.isSatisfiable(Not(e)) match {
          case Some(true) => {println("Invalid"); verified = false}
          case Some(false) => println("Valid")
          case None => {println("Unknown"); verified = false}
        }
      }
      if (verified) println("Program verification successful!")
      else println("Program verification failed")
    }
  }
}
