package lazabs.ai

import lazabs.digraph.Vertex
import lazabs.cfg.{CFG,CFGLabel}
import lazabs.lattice.Lattice
import lazabs.viewer.ScalaPrinter

abstract class TransferFunction[L] {
  // L is the type of abstract domain elements, which a propagated through CFG nodes
  def apply(l : CFGLabel, x : L) : L
  // if x describes a set of states,
  // then apply(n,x) describes states obtained from x by executing statement n
}
 
class AnalysisAlgoritm[L]
               (transferFun : TransferFunction[L],
                lattice : Lattice{type Elem=L},
                cfg : CFG)
{
  var facts : Map[Vertex, L] = Map[Vertex,L]().withDefaultValue(lattice.bottom)
 
  // set particular node
  def setNodeFact(v : Vertex, fact : L) = {facts = facts.updated(v, fact)}
 
  //A naive fixpoint algorithm. Modify if performance is a problem.
  def computeFixpoint : Unit = {
    var change = true 
    while (change) { // iterate until information stabilizes
      change = false
      for (v <- cfg.getVertices) { // process each CFG node
        val oldFact : L = facts(v)   // we had this before
        var newFact : L = oldFact    // start from this
        for ((from,l) <- cfg.in(v)) {  // from --l-> v
          val propagated = transferFun(l, facts(from)) // new info from edge
          newFact = lattice.join(newFact, propagated)      // join new info
        }
        if (newFact != oldFact) {
          change = true;
          facts = facts.updated(v, newFact)
        }
      }
    }
  }
  def getResult : Map[Vertex,L] = facts
}
